#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>



#include "include.h"
#include "device.h"
#include "cloud.h"
#include "edge.h"
#include "pso.h"
#include "ga.h"

#define GA_POP 1000
double chrom[GA_POP][MAX_TASKS];

int chrom_bin[GA_POP][MAX_TASKS];

double fit[GA_POP];

double best_fit = 0.0;
double best_chrom[MAX_TASKS];
int best_chrom_bin[MAX_TASKS];

double max_val[MAX_TASKS];
int core_num[MAX_TASKS];
int chrom_len;

#define MAX_GEN 1000

double prob_mu = 0.1;
double prob_cross = 0.5;
double prob_sel[GA_POP];

void init_gene(int (*cal_fit)(double code[MAX_TASKS]))
{
	int i, j, best = 0;
	srand((unsigned)time(NULL));
	best_fit = 0.0;
	for(i=0; i<GA_POP; i++)
	{
		for(j=0; j<num_task; j++)
		{
			chrom[i][j] = max_val[j] * rand() / RAND_MAX;//0 ~ code_max-1
		}
		fit[i] = 1.0 * cal_fit(chrom[i]);
		if(best_fit < fit[i])
		{
			best_fit = fit[i];
			best = i;
		}
	}
	memcpy(best_chrom, chrom[best], num_task * sizeof(double));
}


void init_gene_bin(int (*cal_fit)(int code[MAX_TASKS]))
{
	int i, j, best = 0;
	srand((unsigned)time(NULL));
	best_fit = 0.0;
	for(i=0; i<GA_POP; i++)
	{
		for(j=0; j<num_task; j++)
		{
			chrom_bin[i][j] = (int)(core_num[j] * rand() / RAND_MAX);//0 ~ code_max-1
			if(chrom_bin[i][j] > core_num[j] - 1)
				chrom_bin[i][j] = core_num[j] - 1;
		}
		fit[i] = 1.0 * cal_fit(chrom_bin[i]);
		if(best_fit < fit[i])
		{
			best_fit = fit[i];
			best = i;
		}
	}
	memcpy(best_chrom_bin, chrom_bin[best], num_task * sizeof(int));
}


void cal_prob_select()
{
	int i;
	memset(prob_sel, 0, GA_POP * sizeof(double));
	prob_sel[0] = fit[0];
	for(i=1; i<GA_POP; i++)
	{
		prob_sel[i] = prob_sel[i-1] + fit[i];
	}
}

int select_one_pop()
{
	int i;
	double rand_value;
	rand_value = prob_sel[GA_POP - 1] * rand() / RAND_MAX;
	if(rand_value <= prob_sel[0])
		return 0;
	for(i=1; i<GA_POP; i++)
	{
		if(prob_sel[i-1] <= rand_value && rand_value <= prob_sel[i])
		{
			return i;
		}
	}
	return -1;
}

int select_another_pop(int one)
{
	int i;
	double rand_value;
	rand_value = (prob_sel[GA_POP - 1] - prob_sel[one]) * rand() / RAND_MAX;
	if(one && rand_value <= prob_sel[0])
		return 0;
	for(i=1; i<one; i++)
	{
		if(prob_sel[i-1] <= rand_value && rand_value <= prob_sel[i])
		{
			return i;
		}
	}
	rand_value += prob_sel[one];
	for(i=one + 1; i<GA_POP; i++)
	{
		if(prob_sel[i-1] <= rand_value && rand_value <= prob_sel[i])
		{
			return i;
		}
	}
	return -1;
}

//one-point crossover
void crossover(int (*cal_fit)(double code[MAX_TASKS]))
{
	int par1, par2;
	int cros;
	double fit0, fit1;
	double child[2][MAX_TASKS];
	par1 = select_one_pop();
	par2 = select_another_pop(par1);
	cros = (int)(1.0 * (num_task - 1) * rand() / RAND_MAX) + 1;//1 ~ num_tasks - 1
	if(cros > num_task - 1)
		cros = num_task - 1;

	memcpy(child[0], chrom[par1], cros * sizeof(double));
	memcpy(child[1], chrom[par2], cros * sizeof(double));
	memcpy(child[0] + cros, chrom[par2] + cros, (num_task - cros) * sizeof(double));
	memcpy(child[1] + cros, chrom[par1] + cros, (num_task - cros) * sizeof(double));
	fit0 = cal_fit(child[0]);
	fit1 = cal_fit(child[1]);
	if(fit0 > fit[par1])
	{
		fit[par1] = fit0;
		memcpy(chrom[par1], child[0], num_task * sizeof(double));
	}
	if(fit0 > fit[par2])
	{
		fit[par2] = fit0;
		memcpy(chrom[par2], child[0], num_task * sizeof(double));
	}
	if(fit1 > fit[par1])
	{
		fit[par1] = fit1;
		memcpy(chrom[par1], child[1], num_task * sizeof(double));
	}
	if(fit1 > fit[par2])
	{
		fit[par2] = fit1;
		memcpy(chrom[par2], child[1], num_task * sizeof(double));
	}

	
	if(best_fit < fit[par1])
	{
		best_fit = fit[par1];
		memcpy(best_chrom, chrom[par1], num_task * sizeof(double));
	}
	if(best_fit < fit[par2])
	{
		best_fit = fit[par2];
		memcpy(best_chrom, chrom[par2], num_task * sizeof(double));
	}
}
	

void exchang(int (*cal_fit)(double code[MAX_TASKS]))
{
	int par;
	double fit0;
	int gene_tmp;
	int gene1, gene2;
	par = select_one_pop();
	gene1 = (int)(1.0 * num_task * rand() / RAND_MAX);//0 ~ num_tasks - 1
	gene2 = (int)(1.0 * num_task * rand() / RAND_MAX);//0 ~ num_tasks - 1
	if(gene1 >= num_task )
		gene1 = num_task - 1;
	if(gene2 >= num_task)
		gene2 = num_task - 1;
	if(gene1 == gene2)
		gene2++;
	gene_tmp = chrom[par][gene1];
	chrom[par][gene1] = chrom[par][gene2];
	chrom[par][gene2] = gene_tmp;
	fit0 = cal_fit(chrom[par]);
	if(fit0 <= fit[par])
	{
		gene_tmp = chrom[par][gene1];
		chrom[par][gene1] = chrom[par][gene2];
		chrom[par][gene2] = gene_tmp;
	}
	else
	{
		fit[par] = fit0;
	}
	
	if(best_fit < fit[par])
	{
		best_fit = fit[par];
		memcpy(best_chrom, chrom[par], num_task * sizeof(double));
	}
}


void mutation(int (*cal_fit)(double code[MAX_TASKS]))
{
	int par;
	double fit0;
	int gene;
	int gene_tmp;
	
	par = select_one_pop();
	
	gene = (int)(1.0 * num_task * rand() / RAND_MAX);//0 ~ num_tasks - 1
	if(gene >= num_task)
		gene = num_task - 1;
	
	gene_tmp = chrom[par][gene];
	chrom[par][gene] = max_val[par] * rand() / RAND_MAX;//0 ~ max_val

	fit0 = cal_fit(chrom[par]);
	if(fit0 <= fit[par])
	{
		chrom[par][gene] = gene_tmp;
	}
	else
	{
		fit[par] = fit0;
	}
	if(best_fit < fit[par])
	{
		best_fit = fit[par];
		memcpy(best_chrom, chrom[par], num_task * sizeof(double));
	}
}


int ga_core()
{
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		max_val[i] = tasks[i].dev->num_core + edge_cores + 1;
	}
	init_gene(cal_fit_core);
	for(i=0; i<MAX_GEN; i++)
	{
		cal_prob_select();
		crossover(cal_fit_core);
//		exchang(cal_fit_core);
		mutation(cal_fit_core);
	}
	
	finish_size = 0.0;
	return cal_fit_core(best_chrom);
}
	

//one-point crossover
void crossover_bin(int (*cal_fit)(int code[MAX_TASKS]))
{
	int par1, par2, i, j;
	int cros;
	int sum_bin = 0;
	double fit0, fit1;
	int child[2][MAX_TASKS];
	par1 = select_one_pop();
	par2 = select_another_pop(par1);
	cros = (int)(1.0 * chrom_len * rand() / RAND_MAX) + 1;//1 ~ chrom_len
	if(cros > chrom_len)
		cros = chrom_len;
	for(i=0; i<num_task; i++)
	{
		sum_bin += core_num[i];
		if(cros >= sum_bin)
		{
			child[0][i] = chrom_bin[par1][i];
			child[1][i] = chrom_bin[par2][i];
		}
		else
			break;
	}
	int exch_bit = (int)(log(sum_bin - cros) / log(2));
	int zero_one = 0x1;
	for(j=0; j<exch_bit; j++)
	{
		zero_one <<= 1;
		zero_one |= 0x1;
	}
	child[0][i] = (chrom_bin[par1][i] | (~zero_one)) & (chrom_bin[par2][i] | (zero_one));
	child[1][i] = (chrom_bin[par2][i] | (~zero_one)) & (chrom_bin[par1][i] | (zero_one));
	for(i++; i<num_task; i++)
	{
		child[0][i] = chrom_bin[par2][i];
		child[1][i] = chrom_bin[par1][i];
	}
	
	fit0 = cal_fit(child[0]);
	fit1 = cal_fit(child[1]);
	
	if(fit0 > fit[par1])
	{
		fit[par1] = fit0;
		memcpy(chrom_bin[par1], child[0], num_task * sizeof(int));
	}
	if(fit0 > fit[par2])
	{
		fit[par2] = fit0;
		memcpy(chrom_bin[par2], child[0], num_task * sizeof(int));
	}
	if(fit1 > fit[par1])
	{
		fit[par1] = fit1;
		memcpy(chrom_bin[par1], child[1], num_task * sizeof(int));
	}
	if(fit1 > fit[par2])
	{
		fit[par2] = fit1;
		memcpy(chrom_bin[par2], child[1], num_task * sizeof(int));
	}

	
	if(best_fit < fit[par1])
	{
		best_fit = fit[par1];
		memcpy(best_chrom_bin, chrom_bin[par1], num_task * sizeof(int));
	}
	if(best_fit < fit[par2])
	{
		best_fit = fit[par2];
		memcpy(best_chrom_bin, chrom_bin[par2], num_task * sizeof(int));
	}
}
	

void mutation_bin(int (*cal_fit)(int code[MAX_TASKS]))
{
	int par1, i, j;
	int mut1, mut2;
	int sum_bin = 0;
	int mut_bit;
	double fit0;
	int child[MAX_TASKS];
	par1 = select_one_pop();
	mut1 = (int)(1.0 * chrom_len * rand() / RAND_MAX) + 1;//1 ~ chrom_len
	if(mut1 > chrom_len)
		mut1 = chrom_len;
	mut2 = (int)(1.0 * chrom_len * rand() / RAND_MAX) + 1;//1 ~ chrom_len
	if(mut2 > chrom_len)
		mut2 = chrom_len;
	while(mut1 == mut2)
	{
		mut2 = (int)(1.0 * chrom_len * rand() / RAND_MAX) + 1;//1 ~ chrom_len
		if(mut2 > chrom_len)
			mut2 = chrom_len;
	}

	sum_bin = 0;
	for(i=0; i<num_task; i++)
	{
		sum_bin += core_num[i];
		if(mut1 >= sum_bin)
		{
			child[i] = chrom_bin[par1][i];
		}
		else
			break;
	}
	mut_bit = (int)(log(sum_bin - mut1) / log(2));
	int zero_one = 0x1;
	for(j=0; j<mut_bit; j++)
	{
		zero_one <<= 1;
	}
	if(child[i] & zero_one)
		child[i] = child[i] | zero_one;
	else
		child[i] = child[i] & ~zero_one;
	for(i++; i<num_task; i++)
	{
		child[i] = chrom_bin[par1][i];
	}
	
	fit0 = cal_fit(child);
	
	if(fit0 > fit[par1])
	{
		fit[par1] = fit0;
		memcpy(chrom_bin[par1], child, num_task * sizeof(int));
	}
	if(best_fit < fit[par1])
	{
		best_fit = fit[par1];
		memcpy(best_chrom_bin, chrom_bin[par1], num_task * sizeof(int));
	}

	sum_bin = 0;
	for(i=0; i<num_task; i++)
	{
		sum_bin += core_num[i];
		if(mut2 >= sum_bin)
		{
			child[i] = chrom_bin[par1][i];
		}
		else
			break;
	}
	mut_bit = (int)(log(sum_bin - mut2) / log(2));
	zero_one = 0x1;
	for(j=0; j<mut_bit; j++)
	{
		zero_one <<= 1;
	}
	if(child[i] & zero_one)
		child[i] = child[i] | zero_one;
	else
		child[i] = child[i] & ~zero_one;
	for(i++; i<num_task; i++)
	{
		child[i] = chrom_bin[par1][i];
	}
	
	fit0 = cal_fit(child);
	
	if(fit0 > fit[par1])
	{
		fit[par1] = fit0;
		memcpy(chrom_bin[par1], child, num_task * sizeof(int));
	}
	if(best_fit < fit[par1])
	{
		best_fit = fit[par1];
		memcpy(best_chrom_bin, chrom_bin[par1], num_task * sizeof(int));
	}
}



int ga_core_bin()
{
	int i, j;
	int edge_cores = 0;
	chrom_len = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		core_num[i] = tasks[i].dev->num_core + edge_cores + 1;
		chrom_len += core_num[i];
	}
	init_gene_bin(cal_fit_core_int);
	for(i=0; i<MAX_GEN; i++)
	{
		cal_prob_select();
		crossover_bin(cal_fit_core_int);
//		exchang(cal_fit_core_int);
		mutation_bin(cal_fit_core_int);
	}
	
	finish_size = 0.0;
	return cal_fit_core_int(best_chrom_bin);
}

