/**
** Particle Swarm Optimization
**/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>



#include "include.h"
#include "device.h"
#include "cloud.h"
#include "edge.h"
#include "pso.h"


double update_pops_range(int (*cal_fit)(double code[MAX_TASKS]))
{
	int i, j;
	double best = 0;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].vel[j] = weight * pops[i].vel[j] 
							+ ac_l * (pops[i].pbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX
							+ ac_g * (gbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX;
			if(pops[i].vel[j] < -code_max)
				pops[i].vel[j] = -code_max;
			else if(pops[i].vel[j] > code_max)
				pops[i].vel[j] = code_max;
			pops[i].code[j] += pops[i].vel[j];

			if(pops[i].code[j] < 0)
				pops[i].code[j] = 0;
			else if(pops[i].code[j] > code_max)
				pops[i].code[j] = code_max;
		}
		best =  1.0 * cal_fit(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}

int pso_range(int (*cal_fit)(double code[MAX_TASKS]))
{
	memset(pops, 0, MAX_POP * sizeof(struct particle));
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop(cal_fit);
	
	int i;
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_range(cal_fit);
//		printf("%d: %.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit(gbest_code);

	return (int)result;
}

int pso_tier_range()
{
	code_max = 3.0;
	code_min = 0.0;
	
	return pso_range(cal_fit_tier);
}

int pso_center_range()
{
	code_max = 2 + num_edge;
	code_min = 0.0;
	return pso_range(cal_fit_center);
}

int pso_srv_range()
{
	code_max = 2.0;
	code_min = 0.0;
	int i;
	for(i=0; i<num_edge; i++)
	{
		code_max += edges[i].num_srv;
	}
	return pso_range(cal_fit_srv);
}

double update_pops_core_range(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].vel[j] = weight * pops[i].vel[j] 
							+ ac_l * (pops[i].pbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX
							+ ac_g * (gbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX;
			pops[i].code[j] += pops[i].vel[j];

			if(pops[i].vel[j] < -code_max[j])
				pops[i].vel[j] = -code_max[j];
			else if(pops[i].vel[j] > code_max[j])
				pops[i].vel[j] = code_max[j];
			pops[i].code[j] += pops[i].vel[j];

			if(pops[i].code[j] < code_min[j])
				pops[i].code[j] = code_min[j];
			else if(pops[i].code[j] > code_max[j])
				pops[i].code[j] = code_max[j];
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}

int pso_core_range()
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	memset(code_max, 0, sizeof(double));
	memset(code_min, 0, sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
	}

	memset(pops, 0, MAX_POP * sizeof(struct particle));
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core(code_max, code_min);
	
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core_range(code_max, code_min);
//		printf("%d: %.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);

	return (int)result;
}

double update_pops_core_dis_range(double code_max[MAX_TASKS], double code_min[MAX_TASKS], double *vel[MAX_POP][MAX_TASKS])
{
	int i, j, k;
	double best = 0;
	double r;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			for(k=0; k<code_max[j]; k++)
			{
				vel[i][j][k] = weight * vel[i][j][k] 
							+ ac_l * dist(pops[i].pbest_code[j], pops[i].code[j])
							+ ac_g * dist(gbest_code[j], pops[i].code[j]);
				if(vel[i][j][k] < -v_max)
					vel[i][j][k] = -v_max;
				else if(vel[i][j][k] > v_max)
					vel[i][j][k] = v_max;
			}
			prob[0] = 1 / (1 + exp(-vel[i][j][0]));
			for(k=1; k<code_max[j]; k++)
			{
				prob[k] = prob[k-1] + 1 / (1 + exp(-vel[i][j][k]));
			}
			r = prob[k-1] * rand() / RAND_MAX;
			if(r <= prob[0])
			{
				pops[i].code[j] = 0.1;
			}
			for(k=0; k<code_max[j]; k++)
			{
				if(prob[k-1] <= r && r <= prob[k])
				{
					pops[i].code[j] = k + 0.1;
				}
			}
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}




/** The method presented in .
**  Zhang Miao, Peng Yong, Yang Mei, Yin Quanjun, Xie Xu,
**  A discrete PSO-based static load balancing algorithm for distributed simulations in a cloud environment,
**  Future Generation Computer Systems,
**  Volume 115,
**  2021,
**  Pages 497-516,
**  ISSN 0167-739X,
**  https://doi.org/10.1016/j.future.2020.09.016.
**/
int pso_core_dis_range()
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	double *vel[MAX_POP][MAX_TASKS];
	memset(code_min, 0, MAX_TASKS * sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
		for(j=0; j<num_pop; j++)
			vel[j][i] = (double *)malloc(code_max[i] * sizeof(double));
	}

	prob = (double *)malloc((MAX_EDGE * MAX_SRV * MAX_CORE + 2) * sizeof(double));
	memset(pops, 0, MAX_POP * sizeof(struct particle));
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core_dis(code_max, code_min, vel);
	
	weight = w_max;
	att = att_min;
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core_dis_range(code_max, code_min, vel);
		att = att_max - (att_max - att_min) * pow(1 - 1.0 * i / max_iter, 5);
//		printf("%d: %.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);
	for(i=0; i<num_task; i++)
	{
		for(j=0; j<num_pop; j++)
			free(vel[j][i]);
	}
	free(prob);

	return (int)result;
}


