#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cloud.h"
#include "edge.h"
#include "device.h"
#include "include.h"
#include "methods.h"
#include "pso.h"
#include "ipso.h"
#include "qpso.h"
#include "ga.h"

struct cloud tmp_clouds[MAX_CLOUD];
int tmp_num_cloud;

struct edge tmp_edges[MAX_EDGE];
int tmp_num_edge;

struct dev tmp_devs[MAX_DEV];
int tmp_num_dev;

struct task tmp_tasks[MAX_TASKS];
double tmp_taskSize = 0.0; //total task size
int tmp_num_task = 0; //fixed total number of sizes

FILE *fp_convergence = NULL;

void save_paras()
{
	memcpy(tmp_clouds, clouds, MAX_CLOUD * sizeof(struct cloud));
	tmp_num_cloud = num_cloud;
	
	memcpy(tmp_edges, edges, MAX_EDGE * sizeof(struct edge));
	tmp_num_edge = num_edge;
	
	memcpy(tmp_devs, devs, MAX_DEV * sizeof(struct dev));
	tmp_num_dev = num_dev;
	
	memcpy(tmp_tasks, tasks, MAX_TASKS * sizeof(struct task));
	tmp_num_task = num_task;
	tmp_taskSize = taskSize;
}


void read_paras()
{
	memcpy(clouds, tmp_clouds, MAX_CLOUD * sizeof(struct cloud));
	num_cloud = tmp_num_cloud;
	
	memcpy(edges, tmp_edges, MAX_EDGE * sizeof(struct edge));
	num_edge = tmp_num_edge;
	
	memcpy(devs, tmp_devs, MAX_DEV * sizeof(struct dev));
	num_dev = tmp_num_dev;
	
	memcpy(tasks, tmp_tasks, MAX_TASKS * sizeof(struct task));
	num_task = tmp_num_task;
	taskSize = tmp_taskSize;
}

int finish_task = 0;
double finish_size = 0.0;

void print_title()
{
	printf("metric\tnumber\tsize\tdata\tcost_eff\tmakespan\tresource\n");
}

void print_result()
{
	int fin_task = 0;
	int tmp_task = 0;
	double fin_size = 0.0;
	double fin_input = 0.0;
	double makespan = 0.0;
	double cost = 0.0;
	double rsc = 0.0;
	double max_timec;
	int index_dev, index_edge, index_srv, index_core, index_cloud, index_vm;
	struct task *pos_task;
	for(index_dev=0; index_dev<num_dev; index_dev++)
	{
		max_timec = 0.0;
		for(index_core=0; index_core<devs[index_dev].num_core; index_core++)
		{
			pos_task = devs[index_dev].cores[index_core].tasks;
			while(pos_task)
			{
				fin_task ++;
				fin_size += pos_task->size;
				fin_input += pos_task->input;
				pos_task = pos_task->next;
			}
			if(makespan < devs[index_dev].cores[index_core].time_c)
				makespan = devs[index_dev].cores[index_core].time_c;
			if(max_timec < devs[index_dev].cores[index_core].time_c)
				max_timec = devs[index_dev].cores[index_core].time_c;
			tmp_task += devs[index_dev].cores[index_core].num_task;
		}
		rsc += max_timec * devs[index_dev].num_core * devs[index_dev].ghz;
	}
	for(index_edge=0; index_edge<num_edge; index_edge++)
	{
		for(index_srv=0; index_srv<edges[index_edge].num_srv; index_srv++)
		{
			max_timec = 0.0;
			for(index_core=0; index_core<edges[index_edge].srv[index_srv].num_core; index_core++)
			{
				pos_task = edges[index_edge].srv[index_srv].cores[index_core].tasks;
				while(pos_task)
				{
					fin_task ++;
					fin_size += pos_task->size;
					fin_input += pos_task->input;
					pos_task = pos_task->next;
				}
				tmp_task += edges[index_edge].srv[index_srv].cores[index_core].num_task;
				if(makespan < edges[index_edge].srv[index_srv].cores[index_core].time_c)
					makespan = edges[index_edge].srv[index_srv].cores[index_core].time_c;
				
				if(max_timec < edges[index_edge].srv[index_srv].cores[index_core].time_c)
					max_timec = edges[index_edge].srv[index_srv].cores[index_core].time_c;
			}
			rsc += max_timec * edges[index_edge].srv[index_srv].num_core * edges[index_edge].srv[index_srv].ghz;
		}
	}
	for(index_cloud=0; index_cloud<num_cloud; index_cloud++)
	{
		for(index_vm=0; index_vm<clouds[index_cloud].num_vm; index_vm++)
		{
			pos_task = clouds[index_cloud].vms[index_vm]->tasks;
			while(pos_task)
			{
				fin_task ++;
				fin_size += pos_task->size;
				fin_input += pos_task->input;
				pos_task = pos_task->next;
			}
			tmp_task += clouds[index_cloud].vms[index_vm]->num_task;
			if(makespan < clouds[index_cloud].vms[index_vm]->time_c)
				makespan = clouds[index_cloud].vms[index_vm]->time_c;
			rsc += clouds[index_cloud].vms[index_vm]->time_c * clouds[index_cloud].ghz;
			cost += ceil(clouds[index_cloud].vms[index_vm]->time_c);
			if(clouds[index_cloud].vms[index_vm])
				free(clouds[index_cloud].vms[index_vm]);
			clouds[index_cloud].vms[index_vm] = NULL;
		}
		clouds[index_cloud].num_vm = 0;
	}
//	printf("finished task number: %d, %.0lf\n", finish_task, finish_size);
//	printf("finished task number: %d, size: %.0lf\n", fin_task, fin_size);
//	printf("finished task number: %d \n", tmp_task);
//	print_title();
	printf("%d\t%.0lf\t%.0lf\t%.2lf\t%lf\t%.0lf\n", 
		fin_task, fin_size, fin_input, cost/fin_size * 1000, makespan, rsc);
}

int main()
{
	char filename[100];
	init_cloud();
	init_edge();
	init_dev();
	init_tsks();
	save_paras();
	printf("%d devices, (%d, %.2lf, %.2lf) tasks, %d edges, %d clouds\n",
			num_dev, num_task, taskSize, total_input, num_edge, num_cloud);
	print_title();

	finish_task = ff();
	printf("FF\t");
	print_result();
	
	read_paras();
	finish_task = ffd();
	printf("FFD\t");
	print_result();

	read_paras();
	finish_task = edf();
	printf("EDF\t");
	print_result();	

	read_paras();
	finish_task = eftf();
	printf("EFTF\t");
	print_result();	

	read_paras();
	finish_task = lact();
	printf("LACT\t");
	print_result();

	
	read_paras();
	finish_task = lstf();
	printf("LSTF\t");
	print_result();	


/*
	read_paras();
	finish_size = 0.0;
	finish_task = pso_tier();
	printf("PSO_tier\t");
	print_result();	

	read_paras();
	finish_size = 0.0;
	finish_task = pso_tier_range();
	printf("PSO_tier_range\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_center();
	printf("PSO_center\t");
	print_result();	

	read_paras();
	finish_size = 0.0;
	finish_task = pso_center_range();
	printf("PSO_center_range\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_srv();
	printf("PSO_srv\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_srv_range();
	printf("PSO_srv_range\t");
	print_result();	
	*/
	for(num_pop = 10; num_pop < 51; num_pop = num_pop + 10) //max -- 0.1
	{
		printf("num_pop: %d \n", num_pop);
	sprintf(filename, "CONVER.PSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_core();
	printf("PSO_core\t");
	print_result();	

	fclose(fp_convergence);

	/*
	//analysis the impact of contraction-expansion coefficient on QPSO
	//
	double cec_max, cec_min;
	for(cec_max = 0.1; cec_max < 2.1; cec_max = cec_max + 0.1) // 2--0.1
	{
		for(cec_min = 0.1; cec_min < cec_max + 0.01; cec_min = cec_min + 0.1) //max -- 0.1
		{
			read_paras();
			finish_size = 0.0;
			qpso_core(qpos_update, LINEAR, cec_min, cec_max);
			printf("QPSO\t");
			printf("[ %.1lf, %.1f ]\t", cec_min, cec_max);
			print_result();	
		}
	}

	for(max_iter = 10; max_iter < 1001; max_iter = max_iter + 10) //max -- 0.1
	{
		read_paras();
		finish_size = 0.0;
		qpso_core(qpos_update, LINEAR, 0.5, 1.0);
		printf("QPSO\t");
		printf("[ %d ]\t", max_iter);
		print_result(); 
	}
*/


///*
	
	sprintf(filename, "CONVER.PSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(qpos_update, LINEAR, 0.5, 1.0);
	printf("QPSO\t");
	print_result();	

	fclose(fp_convergence);
	
	sprintf(filename, "CONVER.HWQPSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(hwqpos_update, LINEAR, 0.5, 1.0);
	printf("HWQPSO\t");
	print_result();	

	fclose(fp_convergence);

	sprintf(filename, "CONVER.Tournament.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	tour_qpso_core(tour_gpso_update, tournament, 0.5, 1.0);
	printf("Tournament\t");
	print_result();	

	fclose(fp_convergence);

	sprintf(filename, "CONVER.GQPSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(gqpos_update, LINEAR, 0.5, 1.0);
	printf("GQPSO\t");
	print_result();	

	fclose(fp_convergence);
	
	sprintf(filename, "CONVER.QPSO_Cauchy.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(qpos_cauchy_update, LINEAR, 0.5, 1.0);
	printf("QPSO_Cauchy\t");
	print_result();	

	fclose(fp_convergence);

	//QIP2020-23-QPSO-CD quantum-behaved particle swarm optimization algorithm with Cauchy distribution
	sprintf(filename, "CONVER.QPSO_CD.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(qpos_cd_update, LINEAR, 0.5, 1.0);
	printf("QPSO_CD\t");
	print_result();	

	fclose(fp_convergence);


	sprintf(filename, "CONVER.AQPSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	//Adaptive quantum behaved particle swarm optimization
	//2005-9-Adaptive Parameter Selection of Quantum-Behaved Particle Swarm Optimization on Global Level
	//parameters: CJ-ExpSys2021-25-Development of some techniques for solving system of linear and nonlinear equations via hybrid algorithm
	qpso_core(qpos_update, ADAPTIVE, 0.5, 1.5);
	printf("AQPSO\t");
	print_result();	

	fclose(fp_convergence);

	sprintf(filename, "CONVER.AGQPSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	//parameters: CJ-ExpSys2021-25-Development of some techniques for solving system of linear and nonlinear equations via hybrid algorithm
	qpso_core(gqpos_update, ADAPTIVE, 0.5, 1.5);
	printf("AGQPSO\t");
	print_result();

	fclose(fp_convergence);

	sprintf(filename, "CONVER.LQPSO.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	//parameters: CJ-ExpSys2021-25-Development of some techniques for solving system of linear and nonlinear equations via hybrid algorithm
	qpso_core(lqpos_update, QUAD_DEC, 0.5, 1.0);
	printf("LQPSO\t");
	print_result();	


	fclose(fp_convergence);
	sprintf(filename, "CONVER.QPSO_Quad.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(qpos_update, QUADRATIC, 0.5, 1.0);
	printf("QPSO_Quad\t");
	print_result();	


	fclose(fp_convergence);
	sprintf(filename, "CONVER.QPSO_Quad_DEC.%dPOP.txt", num_pop);
	fp_convergence = fopen(filename, "w");

	read_paras();
	finish_size = 0.0;
	qpso_core(qpos_update, QUAD_DEC, 0.5, 1.0);
	printf("QPSO_Quad_DEC\t");
	print_result();	

	fclose(fp_convergence);
	}
	//*/	
/*	read_paras();
	finish_size = 0.0;
	finish_task = pso_core_range();
	printf("PSO_core_range\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_core_dis();
	printf("pso_core_dis\t");
	print_result();	


	read_paras();
	finish_size = 0.0;
	finish_task = pso_core_dis_range();
	printf("pso_core_dis_range\t");
	print_result();	

	read_paras();
	finish_size = 0.0;
	finish_task = ga_core();
	printf("ga_core\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = ga_core_bin();
	printf("ga_core_bin\t");
	print_result();	
	
	read_paras();
	finish_size = 0.0;
	finish_task = pso_core_bin();
	printf("pso_core_bin\t");
	print_result();	

	read_paras();
	finish_size = 0.0;
	finish_task = pso_core_map();
	printf("pso_core_map\t");
	print_result();	*/
	return 0;
}
