/**
** Particle Swarm Optimization
**/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>



#include "include.h"
#include "device.h"
#include "cloud.h"
#include "edge.h"
#include "pso.h"


int num_pop = 50;// 500
int max_iter = 100; // 1000

double code_max = 3.0;
double code_min = 0.0;
//vel = -code_max ~ code_max

struct particle pops[MAX_POP];


double gbest_code[MAX_TASKS]; //the code providing gbest
double gbest = 0.0;

//pso_bin
int gb_code_bin[MAX_TASKS];
int gb_bin;
int pb_code_bin[MAX_POP][MAX_TASKS];
int pb_bin[MAX_POP];
int core_num[MAX_TASKS];
int code[MAX_POP][MAX_TASKS];
int code_len[MAX_TASKS];
double vel[MAX_POP][MAX_TASKS][32];


//pso_bin
double vel_map[MAX_POP][MAX_TASKS][MAX_CORES];
char code_map[MAX_POP][MAX_TASKS][MAX_CORES];
char gb_code_map[MAX_TASKS][MAX_CORES];
int gb_map;
char pb_code_map[MAX_POP][MAX_TASKS][MAX_CORES];
int pb_map[MAX_POP];

double w_max = 1.4; //Inertia Weight :1.4~0
double w_min = 0;
double weight;

//Acceleration Coefficients
double ac_p = 2.0;
double ac_g = 2.0;
double ac_l = 2.0;
int test_task;

//for pso_core_dis()
double att = 0;
double att_max = 5;
double att_min = 0;
double v_max = 5;
double *prob = 0;
int sch_core(struct task *tsk, struct core *pcore, double ghz, double bw_down)
{
	struct task *pos;
	double time_c;
	double time_i;
	
	time_c = 0.0;
	time_i = 0.0;
	pos = pcore->tasks;
	while(pos && pos->deadline < tsk->deadline)
	{
		if(bw_down > 1e-6)
			time_i += pos->input / bw_down;
		if(time_c > time_i)
			time_c += pos->size / ghz;
		else
			time_c = time_i + pos->size / ghz;
		
		if(time_c > pos->deadline)
		{
			return 0;
		}
		pos = pos->next;
	}
	
	if(bw_down > 1e-6)
		time_i += tsk->input / bw_down;
	if(time_c > time_i)
		time_c += tsk->size / ghz;
	else
		time_c = time_i + tsk->size / ghz;
	if(time_c > tsk->deadline)
		return 0;
	
	while(pos)
	{
		if(bw_down > 1e-6)
			time_i += pos->input / bw_down;
		if(time_c > time_i)
			time_c += pos->size / ghz;
		else
			time_c = time_i + pos->size / ghz;
		if(time_c > pos->deadline)
		{
			return 0;
		}
		pos = pos->next;
	}
	pcore->time_c = time_c;
	pcore->time_in = time_i;
	pcore->num_task++;
	finish_size += tsk->size;
	insert_dl_inc(&pcore->tasks, tsk, &pcore->tail);
	
	return 1;
}
int sch_dev(struct task *tsk)
{
	struct dev *pdev = tsk->dev;
	int i;
	for(i=0; i<pdev->num_core; i++)
	{
		if(sch_core(tsk, pdev->cores + i, pdev->ghz, 0.0))
			return 1;
	}
	return 0;
}
int sch_srv(struct task *tsk, struct edge *pedge, struct server *psrv)
{
	int i;
	for(i=0; i<psrv->num_core; i++)
	{
		if(sch_core(tsk, psrv->cores + i, psrv->ghz, pedge->bw_down))
			return 1;
	}		
	return 0;
}

int sch_edge(struct task *tsk)
{
	struct edge *pedge;
	struct server *psrv;
	int i, j;
	for(i=0; i<num_edge; i++)
	{
		pedge = edges + i;
		for(j=0; j<pedge->num_srv; j++)
		{
			psrv = pedge->srv + j;
			if(sch_srv(tsk, pedge, psrv))
				return 1;
		}
	}

	return 0;
}


int sch_an_edge(struct task *tsk, struct edge *pedge)
{
	struct server *psrv;
	int j;
	for(j=0; j<pedge->num_srv; j++)
	{
		psrv = pedge->srv + j;
		if(sch_srv(tsk, pedge, psrv))
			return 1;
	}

	return 0;
}

int sch_cloud(struct task *tsk)
{
	struct task *pos;
	struct cloud *pcloud;
	struct vm *pvm;
	int i, j;
	int flag = 1;
	double time_c;
	double time_i;
	for(i=0; i<num_cloud; i++)
	{
		pcloud = clouds + i;
		for(j=0; j<pcloud->num_vm; j++)
		{
			pvm = pcloud->vms[j];
			time_c = 0.0;
			time_i = 0.0;
			flag = 1;
			pos = pvm->tasks;
			while(pos && pos->deadline < tsk->deadline)
			{
				time_i += pos->input / pcloud->bw_down;
				if(time_c > time_i)
					time_c += pos->size / pcloud->ghz;
				else
					time_c = time_i + pos->size / pcloud->ghz;
				
				if(time_c > pos->deadline)
				{
					flag = 0;
					break;
				}
				pos = pos->next;
			}
			if(0 == flag)
				continue;
			time_i += tsk->input / pcloud->bw_down;
			if(time_c > time_i)
				time_c += tsk->size / pcloud->ghz;
			else
				time_c = time_i + tsk->size / pcloud->ghz;
			if(time_c > tsk->deadline)
				continue;
			while(pos)
			{
				time_i += pos->input / pcloud->bw_down;
				if(time_c > time_i)
					time_c += pos->size / pcloud->ghz;
				else
					time_c = time_i + pos->size / pcloud->ghz;
				if(time_c > pos->deadline)
				{
					break;
				}
				pos = pos->next;
			}
			if(NULL == pos)
			{
				pvm->time_c = time_c;
				pvm->time_in = time_i;
				pvm->num_task++;
				finish_size += tsk->size;
				insert_dl_inc(&pvm->tasks, tsk, &pvm->tail);
				return 1;
			}
		}		
	}

	//rent a new vm
	for(i=0; i<num_cloud; i++)
	{
		pcloud = clouds + i;
		time_i = tsk->input / pcloud->bw_down;
		time_c = time_i + tsk->size / pcloud->ghz;
		if(time_c > tsk->deadline)
			continue;

		pvm = (struct vm *) malloc(sizeof(struct vm));
		memset(pvm, 0, sizeof(struct vm));
		pcloud->vms[pcloud->num_vm] = pvm;
		pcloud->num_vm++;
		pvm->time_c = time_c;
		pvm->time_in = time_i;
		pvm->num_task++;
		finish_size += tsk->size;
		insert_dl_inc(&pvm->tasks, tsk, &pvm->tail);
		return 1;
	}
	return 0;
}

void clean_status()
{
	int i_dev, i_edge, i_core, i_srv, i_cloud, i_vm;
	for(i_dev=0; i_dev<num_dev; i_dev++)
	{
		for(i_core=0; i_core<devs[i_dev].num_core; i_core++)
		{
			devs[i_dev].cores[i_core].tasks = NULL;
			devs[i_dev].cores[i_core].tail = NULL;
			devs[i_dev].cores[i_core].num_task = 0;
			devs[i_dev].cores[i_core].time_c = 0.0;
		}
	}
	for(i_edge=0; i_edge<num_edge; i_edge++)
	{
		for(i_srv=0; i_srv<edges[i_edge].num_srv; i_srv++)
		{
			for(i_core=0; i_core<edges[i_edge].srv[i_srv].num_core; i_core++)
			{
				edges[i_edge].srv[i_srv].cores[i_core].tasks = NULL;
				edges[i_edge].srv[i_srv].cores[i_core].tail = NULL;
				edges[i_edge].srv[i_srv].cores[i_core].num_task = 0;
				edges[i_edge].srv[i_srv].cores[i_core].time_c = 0.0;
				edges[i_edge].srv[i_srv].cores[i_core].time_in = 0.0;
			}
		}
	}
	for(i_cloud=0; i_cloud<num_cloud; i_cloud++)
	{
		for(i_vm=0; i_vm<clouds[i_cloud].num_vm; i_vm++)
		{
			if(clouds[i_cloud].vms[i_vm])
				free(clouds[i_cloud].vms[i_vm]);
			clouds[i_cloud].vms[i_vm] = NULL;
		}
		clouds[i_cloud].num_vm = 0;
	}
}

int cal_fit_tier(double code[MAX_TASKS])
{
	int i;
	int finish = 0;

	clean_status();
	for(i=0; i<num_task; i++)
	{
		if(code[i] < 1.0)
			finish += sch_dev(tasks + i);
		else if(code[i] < 2.0)
			finish += sch_edge(tasks + i);
		else
			finish += sch_cloud(tasks + i);
	}
	return finish;
}


int cal_fit_center(double code[MAX_TASKS])
{
	int i;
	int finish = 0;

	clean_status();
	for(i=0; i<num_task; i++)
	{
		if(code[i] < 1.0)
			finish += sch_dev(tasks + i);
		else if(code[i] < code_max - 1)
			finish += sch_an_edge(tasks + i, edges + (int)(code[i]) - 1);
		else
		{
			finish += sch_cloud(tasks + i);
		}
	}
	return finish;
}



int cal_fit_srv(double code[MAX_TASKS])
{
	int i, j, isrv;
	int finish = 0;

	clean_status();
	for(i=0; i<num_task; i++)
	{
		if(code[i] < 1.0)
			finish += sch_dev(tasks + i);
		else if(code[i] < code_max - 1)
		{
			isrv = (int)code[i] - 1;
			for(j=0; j<num_edge; j++)
			{
				if(isrv >= edges[j].num_srv)
					isrv -= edges[j].num_srv;
				else
					break;
			}
			finish += sch_srv(tasks + i, edges + j, edges[j].srv + isrv);
		}
		else
		{
			finish += sch_cloud(tasks + i);
		}
	}
	return finish;
}

int cal_fit_core(double code[MAX_TASKS])
{
	int i, j, k;
	int finish = 0;
	int icore;
	struct dev * pdev;
	struct edge *pedge = edges;
	struct server *psrv = pedge->srv;
	struct task *ptsk;

	clean_status();
	for(i=0; i<num_task; i++)
	{
		ptsk = tasks + i;
		pdev = ptsk->dev;
		if(code[i] < pdev->num_core)
			finish += sch_core(ptsk, pdev->cores + (int)code[i], pdev->ghz, 0.0);
		else if(code[i] < code_max - 1)
		{
			icore = (int)code[i] - pdev->num_core;
			for(j=0; j<num_edge; j++)
			{
				pedge = edges + j;
				for(k=0; k<pedge->num_srv; k++)
				{
					psrv = pedge->srv + k;
					if(icore >= psrv->num_core)
						icore -= psrv->num_core;
					else
						goto endFor;
				}
			}
			endFor:
			finish += sch_core(ptsk, psrv->cores + icore, pdev->ghz, pedge->bw_down);
		}
		else
			finish += sch_cloud(ptsk);
	}
	return finish;
}

int cal_fit_core_int(int code[MAX_TASKS])
{
	int i;
	double code_double[MAX_TASKS];
	for(i=0; i<num_task; i++)
	{
		code_double[i] = code[i] - 0.1;
	}
	return cal_fit_core(code_double);
}

int cal_fit_core_bin(int code[MAX_TASKS])
{
	int i;
	double code_double[MAX_TASKS];
	for(i=0; i<num_task; i++)
	{
		code_double[i] = code[i] - 0.1;
	}
	return cal_fit_core(code_double);
}


void init_pop(int (*cal_fit)(double code[MAX_TASKS]))
{
	int i, j;
	srand((unsigned)time(NULL));
	gbest = 0.0;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].code[j] = code_max * rand() / RAND_MAX;//0 ~ code_max
			pops[i].vel[j] = 2 * code_max * rand() / RAND_MAX - code_max; // -code_max ~ code_max
			pops[i].pbest = 1.0 * cal_fit(pops[i].code);
		}
		memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
}
double update_pops(int (*cal_fit)(double code[MAX_TASKS]))
{
	int i, j;
	double best = 0;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].vel[j] = weight * pops[i].vel[j] 
							+ ac_l * (pops[i].pbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX
							+ ac_g * (gbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX;
			pops[i].code[j] += pops[i].vel[j];

			//if code_min < 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max) * code_max - pops[i].code[j]);
		}
		best =  1.0 * cal_fit(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}

int pso(int (*cal_fit)(double code[MAX_TASKS]))
{
	memset(pops, 0, MAX_POP * sizeof(struct particle));
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop(cal_fit);
	
	int i;
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops(cal_fit);
//		printf("%d: %.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit(gbest_code);

	return (int)result;
}

int pso_tier()
{
	code_max = 3.0;
	code_min = 0.0;
	
	return pso(cal_fit_tier);
}

int pso_center()
{
	code_max = 2 + num_edge;
	code_min = 0.0;
	return pso(cal_fit_center);
}

int pso_srv()
{
	code_max = 2.0;
	code_min = 0.0;
	int i;
	for(i=0; i<num_edge; i++)
	{
		code_max += edges[i].num_srv;
	}
	return pso(cal_fit_srv);
}


void init_pop_core(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	srand((unsigned)time(NULL));
	gbest = 0.0;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].code[j] = code_max[j] * rand() / RAND_MAX;//0 ~ code_max
			pops[i].vel[j] = 2 * code_max[j] * rand() / RAND_MAX - code_max[j]; // -code_max ~ code_max
			pops[i].fit = pops[i].pbest = 1.0 * cal_fit_core(pops[i].code);
		}
		memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
}


void init_pop_core_bin()
{
	int i, j, k;
	srand((unsigned)time(NULL));
	gb_bin = 0.0;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			code[i][j] = (int)(core_num[j] * rand() / RAND_MAX);//0 ~ code_max-1
			if(code[i][j] > core_num[j] - 1)
				code[i][j] = core_num[j] - 1;
			for(k=0; k<code_len[j]; k++)
				vel[i][j][k] = 2.0 * rand() / RAND_MAX - 1; // -1 ~ 1
			
			pb_bin[i] = cal_fit_core_bin(code[i]);
		}
		memcpy(pb_code_bin, code, MAX_POP * MAX_TASKS * sizeof(int));
		if(pb_bin[i] > gb_bin)
		{
			gb_bin = pb_bin[i];
			memcpy(gb_code_bin, pb_code_bin[i], MAX_TASKS * sizeof(int));
		}
	}
}

double update_pops_core(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			pops[i].vel[j] = weight * pops[i].vel[j] 
							+ ac_l * (pops[i].pbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX
							+ ac_g * (gbest_code[j] - pops[i].code[j]) * rand() / RAND_MAX;
			pops[i].code[j] += pops[i].vel[j];

			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}


int update_pops_core_bin()
{
	int i, j, k;
	int best = 0;
	int code_bit;
	int pb_code_bit;
	int gb_code_bit;
	int code_tmp;
	int pb_code_tmp;
	int gb_code_tmp;
	weight -= (w_max - w_min) / max_iter;

	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			code_tmp = code[i][j];
			pb_code_tmp = pb_code_bin[i][j];
			gb_code_tmp = gb_code_bin[j];
			
			for(k=0; k<code_len[j]; k++)
			{
				code_bit = code_tmp & 0x01;
				pb_code_bit = pb_code_tmp & 0x01;
				gb_code_bit = gb_code_tmp & 0x01;
				
				vel[i][j][k] = weight * vel[i][j][k]
							+ ac_l * (pb_code_bit - code_bit) * rand() / RAND_MAX
							+ ac_g * (gb_code_bit - code_bit) * rand() / RAND_MAX;

				if(1.0 * rand() / RAND_MAX < 1/ (1 + exp(-vel[i][j][k])))
					code[i][j] |= 0x01 << k;
				else
					code[i][j] &= 0xFFFFFFFF ^ (0x01 << k);
				code_tmp >>= 1;
				pb_code_tmp >>= 1;
				gb_code_tmp >>= 1;
			}
		}
		best =  cal_fit_core_bin(code[i]);
		if(pb_bin[i] < best)
		{
			pb_bin[i] = best;
			memcpy(pb_code_bin[i], code[i], num_task * sizeof(int));
		}
		if(pb_bin[i] > gb_bin)
		{
			gb_bin = pb_bin[i];
			memcpy(gb_code_bin, pb_code_bin[i], num_task * sizeof(int));
		}
	}
	return gb_bin;
}

int pso_core()
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	memset(code_max, 0, sizeof(double));
	memset(code_min, 0, sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
	}

	memset(pops, 0, MAX_POP * sizeof(struct particle));
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core(code_max, code_min);
	
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core(code_max, code_min);
		fprintf(fp_convergence, "%d\t%.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);

	return (int)result;
}



void init_pop_core_dis(double code_max[MAX_TASKS], double code_min[MAX_TASKS], double *vel[MAX_POP][MAX_TASKS])
{
	int i, j, k;
	srand((unsigned)time(NULL));
	gbest = 0.0;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			for(k=0; k<code_max[j]; k++)
				vel[i][j][k] = 2 * v_max * rand() / RAND_MAX - v_max; // -v_max ~ v_max
			pops[i].code[j] = code_max[j] * rand() / RAND_MAX;//0 ~ code_max
			pops[i].pbest = 1.0 * cal_fit_core(pops[i].code);
		}
		memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
}

double dist(double x, double y)
{
	if(fabs(x - y) < 1e-6)
		return att;
	else
		return -att;
}
double update_pops_core_dis(double code_max[MAX_TASKS], double code_min[MAX_TASKS], double *vel[MAX_POP][MAX_TASKS])
{
	int i, j, k;
	double best = 0;
	double r;
	weight -= (w_max - w_min) / max_iter;
//	double vel_max = w_max;
//	double vel_min = w_min;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			for(k=0; k<code_max[j]; k++)
			{
				vel[i][j][k] = weight * vel[i][j][k] 
							+ ac_l * dist(pops[i].pbest_code[j], pops[i].code[j])
							+ ac_g * dist(gbest_code[j], pops[i].code[j]);
			}
			prob[0] = 1 / (1 + exp(-vel[i][j][0]));
			for(k=1; k<code_max[j]; k++)
			{
				prob[k] = prob[k-1] + 1 / (1 + exp(-vel[i][j][k]));
			}
			r = prob[k-1] * rand() / RAND_MAX;
			if(r <= prob[0])
			{
				pops[i].code[j] = 0.1;
			}
			for(k=0; k<code_max[j]; k++)
			{
				if(prob[k-1] <= r && r <= prob[k])
				{
					pops[i].code[j] = k + 0.1;
				}
			}
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return gbest;
}


/** The method presented in .
**  Zhang Miao, Peng Yong, Yang Mei, Yin Quanjun, Xie Xu,
**  A discrete PSO-based static load balancing algorithm for distributed simulations in a cloud environment,
**  Future Generation Computer Systems,
**  Volume 115,
**  2021,
**  Pages 497-516,
**  ISSN 0167-739X,
**  https://doi.org/10.1016/j.future.2020.09.016.
**/
int pso_core_dis()
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	double *vel[MAX_POP][MAX_TASKS];
	memset(code_min, 0, MAX_TASKS * sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
		for(j=0; j<num_pop; j++)
			vel[j][i] = (double *)malloc(code_max[i] * sizeof(double));
	}

	prob = (double *)malloc((MAX_EDGE * MAX_SRV * MAX_CORE + 2) * sizeof(double));
	memset(pops, 0, MAX_POP * sizeof(struct particle));
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core_dis(code_max, code_min, vel);
	
	weight = w_max;
	att = att_min;
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core_dis(code_max, code_min, vel);
		att = att_max - (att_max - att_min) * pow(1 - 1.0 * i / max_iter, 5);
//		printf("%d: %.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);
	for(i=0; i<num_task; i++)
	{
		for(j=0; j<num_pop; j++)
			free(vel[j][i]);
	}
	free(prob);

	return (int)result;
}
int bit_len(int number)
{
	int i = 0;
	while(number)
	{
		i++;
		number >>= 1;
	}
	return i;
}

int pso_core_bin()
{
	int i, j;
	int edge_core = 0;
	int tmp;

	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_core += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		core_num[i] = tasks[i].dev->num_core + edge_core + num_cloud;
		code_len[i] = 0;
		tmp = core_num[i];
		while(tmp)
		{
			code_len[i]++;
			tmp >>= 1;
		}
	}
	for(i=0; i<num_pop; i++)
	{
		pb_bin[i] = 0;
	}	
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	int result;
	
	init_pop_core_bin();
	
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core_bin();
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core_bin(gb_code_bin);
	return result;
}


int cal_fit_core_map(char code[MAX_TASKS][MAX_CORES])
{
	int i, tmp, last;
	double code_double[MAX_TASKS];
	for(i=0; i<num_task; i++)
	{
		last = 0;
		for(tmp=0; tmp<core_num[i]; tmp++)
		{
			if(code[i][tmp])
			{
				last = tmp;
				if((rand()&0x01))
				{
					break;
				}
			}
		}
		code_double[i] = last - 0.1;
	}
	return cal_fit_core(code_double);
}


void init_pop_core_map()
{
	int i, j, k;
	srand((unsigned)time(NULL));
	gb_map = 0.0;
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			for(k=0; k<core_num[j]; k++)
			{
				code_map[i][j][k] = rand() & 0x01; //0 or 1
				vel[i][j][k] = 2.0 * rand() / RAND_MAX - 1; // -1 ~ 1
			}
			
			pb_map[i] = cal_fit_core_map(code_map[i]);
		}
		memcpy(pb_code_map, code_map, MAX_POP * MAX_TASKS * MAX_CORES * sizeof(char));
		if(pb_map[i] > gb_map)
		{
			gb_map = pb_map[i];
			memcpy(gb_code_map, pb_code_map[i], MAX_TASKS * MAX_CORES * sizeof(char));
		}
	}
}


int update_pops_core_map()
{
	int i, j, k;
	int best = 0;
	
	weight -= (w_max - w_min) / max_iter;

	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{			
			for(k=0; k<core_num[j]; k++)
			{				
				vel[i][j][k] = weight * vel[i][j][k]
							+ ac_l * (pb_code_map[i][j][k] - code_map[i][j][k]) * rand() / RAND_MAX
							+ ac_g * (gb_code_map[j][k] - code_map[i][j][k]) * rand() / RAND_MAX;

				if(1.0 * rand() / RAND_MAX < 1/ (1 + exp(-vel[i][j][k])))
					code_map[i][j][k] = 1;
				else
					code_map[i][j][k] = 0;
			}
		}
		best =  cal_fit_core_map(code_map[i]);
		if(pb_map[i] < best)
		{
			pb_map[i] = best;
			memcpy(pb_code_map[i], code_map[i], MAX_TASKS * MAX_CORES * sizeof(char));
		}
		if(pb_map[i] > gb_map)
		{
			gb_map = pb_map[i];
			memcpy(gb_code_map, pb_code_map[i], MAX_TASKS * MAX_CORES * sizeof(char));
		}
	}
	return gb_map;
}


int pso_core_map()
{
	int i, j;
	int edge_core = 0;

	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_core += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		core_num[i] = tasks[i].dev->num_core + edge_core + num_cloud;
	}
	gb_map = 0;
	for(i=0; i<num_pop; i++)
	{
		pb_map[i] = 0;
	}
	weight = w_max;
	struct timeval start, end;
//	double timeuse;
	int result;
	
	init_pop_core_map();
	
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		result = update_pops_core_map();
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core_map(gb_code_map);
	return result;
}

