#ifndef _PSO_H_
#define _PSO_H_

#include "task.h"

struct particle{
	double code[MAX_TASKS]; //the position vector. for each task, 0 for local (0~1), 1 for edge (1~2), and 2 for cloud (2~3)
	double fit;
	double vel[MAX_TASKS]; //velocity

	double center[MAX_TASKS]; //local attractor for QPSO
	
	double pbest_code[MAX_TASKS];  //the code providing pbest
	double pbest;
	
	//for Local Version PSO
	double local_best[MAX_TASKS];
	double fit_lbest[MAX_TASKS];

};

#define MAX_POP 200

extern struct particle pops[MAX_POP];

extern double code_max;
extern double code_min;


extern int num_pop;
extern int max_iter;

extern double gbest_code[MAX_TASKS];
extern double gbest;

extern int gb_code_bin[MAX_TASKS];
extern int gb_bin;

extern char gb_code_map[MAX_TASKS][MAX_CORES];
extern int gb_map;



extern double w_max;
extern double w_min;
extern double weight;
extern double ac_p;
extern double ac_g;
extern double ac_l;

//for pso_core_dis()
extern double att;
extern double att_max;
extern double att_min;
extern double v_max;
extern double *prob;


int pso_tier();
int pso_center();
int pso_srv();
int pso_core();

int pso_core_dis();
int pso_core_bin();
int pso_core_map();



int cal_fit_tier(double code[MAX_TASKS]);
int cal_fit_center(double code[MAX_TASKS]);
int cal_fit_srv(double code[MAX_TASKS]);
int cal_fit_core(double code[MAX_TASKS]);
int cal_fit_core_int(int code[MAX_TASKS]);


void init_pop(int (*cal_fit)(double code[MAX_TASKS]));
void init_pop_core(double code_max[MAX_TASKS], double code_min[MAX_TASKS]);
void init_pop_core_dis(double code_max[MAX_TASKS], double code_min[MAX_TASKS], double *vel[MAX_POP][MAX_TASKS]);


int sch_core(struct task *tsk, struct core *pcore, double ghz, double bw_down);
int sch_dev(struct task *tsk);
int sch_srv(struct task *tsk, struct edge *pedge, struct server *psrv);
int sch_edge(struct task *tsk);
int sch_an_edge(struct task *tsk, struct edge *pedge);
int sch_cloud(struct task *tsk);


void clean_status();

double dist(double x, double y);


#endif
