#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>



#include "include.h"
#include "device.h"
#include "cloud.h"
#include "edge.h"
#include "pso.h"
#include "qpso.h"

//for qpso
double alpha;//contraction-expansion coefficient
double alpha_max = 1.0;
double alpha_min = 0.5;
double alpha_detal = 0.5;
//for aqpos
//diversity bound
double alpha_bound = 1.7;// Attraction when <= diver_bound, diver_bound else
double diver_low = 0.000005;
double diver_high = 0.25;

//for Tournament AGQPSO
#define NUM_SWARM 4
#define GROUP_SIZE 2
//double tourn_gb_code[MAX_TASKS][NUM_SWARM]; //the code providing gbest
//double tourn_gb[NUM_SWARM];


void init_code_range(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	memset(code_max, 0, sizeof(double));
	memset(code_min, 0, sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
	}
}

//https://blog.csdn.net/weixin_38902950/article/details/91446353
//Box-Muller 
double gaussrand( )
{
	static double U, V;
	static int phase = 0;
	double Z;

	if(phase == 0)
	{
		U = rand() / (RAND_MAX + 1.0);
		V = rand() / (RAND_MAX + 1.0);
		Z = sqrt(-2.0 * log(U))* sin(2.0 * M_PI * V);
	}
	else
	{
		Z = sqrt(-2.0 * log(U)) * cos(2.0 * M_PI * V);
	}

	phase = 1 - phase;
	if(Z < 0)
		Z = -Z;
	return Z;
}

double cauchy_rand( )
{
	double x = rand() / (RAND_MAX + 1.0);
	return 1 / (1 + (x * x)) / M_PI;
}

double lagam(double x)
{
	int i; double y, t, s, u ;
	double a[ 11] = { 0.0000677106, - 0.0003442342, 0.0015397681, - 0.0024467480, 0.0109736958,
		-0.0002109075, 0.0742379071, 0.0815782188, 0.4118402518, 0.4227843370, 1.0};
	if ( x <= 0.0)
	{
		printf("err** x <= 0! \n");
		return ( -1.0);
	}
	y = x;
	if ( y <= 1.0)
	{
		t = 1.0 / ( y * ( y + 1.0) );
		y = y + 2.0;
	}
	else if ( y <= 2.0)
	{
		t = 1.0 / y;
		y = y + 1.0;
	}
	else if ( y <= 3.0)
		t = 1.0;
	else
	{
		t = 1.0;
		while ( y > 3.0)
		{
			y = y - 1.0;
			t = t * y;
		}
	}
	s = a[0];
	u = y - 2.0;
	for( i=1; i <= 10; i++ )
		s = s * u + a[i];
	s = s* t;
	return( s) ;
}



//QPSO with delta potential well
double qpos_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai, c, ln_u;
	
	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pop; i++)
		{
			tmp += pops[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai = 1.0 * rand() / RAND_MAX;
			c = fai * pops[i].pbest_code[j] + (1 - fai) * gbest_code[j];
			ln_u = log(1.0 * RAND_MAX / rand());
			if(rand() < RAND_MAX / 2)
				pops[i].code[j] = c + alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			else
				pops[i].code[j] = c - alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);

			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}

double qpos_update_with_method(double code_max [ MAX_TASKS ], double code_min [ MAX_TASKS ], double(*rand_method)())
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai1, fai2, c, ln_u;
	
	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pop; i++)
		{
			tmp += pops[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai1 = rand_method();
			fai2 = rand_method();
			c = (fai1 * pops[i].pbest_code[j] + fai2 * gbest_code[j]) / (fai1 + fai2);
			ln_u = log(fai1);
			if(rand() < RAND_MAX / 2)
				pops[i].code[j] = c + alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			else
				pops[i].code[j] = c - alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);

			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}


//QPSO with gauss rand
double gqpos_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	return qpos_update_with_method(code_max, code_min, gaussrand);
}
//QPSO with Cauchy distribution
double qpos_cauchy_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	return qpos_update_with_method(code_max, code_min, cauchy_rand);
}
void sort_fit_dec(struct particle *swarm, int num_pcls);

double qpos_cd_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai1, fai2, c, ln_u;
	
	double prob_mu = 0.1;
	
	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pop; i++)
		{
			tmp += pops[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai1 = cauchy_rand();
			fai2 = cauchy_rand();
			c = (fai1 * pops[i].pbest_code[j] + fai2 * gbest_code[j]) / (fai1 + fai2);
			ln_u = log(cauchy_rand());
			if(rand() < RAND_MAX / 2)
				pops[i].code[j] = c + alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			else
				pops[i].code[j] = c - alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);

			if(rand() < prob_mu * RAND_MAX)
			{
				pops[i].code[j] += cauchy_rand() * rand() / (RAND_MAX + 1.0);
			}
			
			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		pops[i].fit = best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
		sort_fit_dec(pops, num_pop);
		memcpy(pops + num_pop / 2, pops, num_pop / 2 * sizeof(struct particle));
	}
	return best;
}



//the length of longest the diagonal in the search space
double cal_diagonal(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	double dia = 0.0;
	int i;
	for(i=0; i<num_pop; i++)
		dia += (code_max[i] - code_min[i]) * (code_max[i] - code_min[i]);
	dia  = sqrt(dia);
	return dia;
}

//the diversity of the swarm referring to
//2005-9-Adaptive Parameter Selection of Quantum-Behaved Particle Swarm Optimization on Global Level
double cal_diversity(struct particle pops[MAX_POP], int num_pop, int pop_size,
						double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double std, diver = 0.0;
	double code_avg[MAX_TASKS];
	for(j=0; j<pop_size; j++)
	{
		code_avg[j] = 0.0;
		for(i=0; i<num_pop; i++)
		{
			code_avg[j] += pops[i].code[j];
		}
		code_avg[j] = code_avg[j] / num_pop;
	}
	for(i=0; i<num_pop; i++)
	{
		std = 0.0;
		for(j=0; j<pop_size; j++)
		{
			std += (pops[i].code[j] - code_avg[j]) * (pops[i].code[j] - code_avg[j]);
		}
		diver += sqrt(std);
	}
	diver = diver / num_pop / cal_diagonal(code_max, code_min);
	return 0.0;
}

//qpos with Lévy flight
//ASC2021-9-QPSO algorithm based on Lévy flight and its application in fuzzy portfolio
double lqpos_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai, c, ln_u, step;

	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pop; i++)
		{
			tmp += pops[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai = 1.0 * rand() / RAND_MAX;
			c = fai * pops[i].pbest_code[j] + (1 - fai) * gbest_code[j];
			ln_u = log(1.0 * RAND_MAX / rand());
			if(ln_u < log(2))
			{
				double k = 2.0 * rand() / RAND_MAX + 1;
				double mu = gaussrand();
				double nu = gaussrand();
				double delta;
				delta = lagam(1 + k) * sin(0.5 * k * M_PI) / lagam((1 + k) * 0.5) / pow(2, (k - 1) * 0.5);
				nu = nu * pow(delta, 1 / 2.0 / k);
				step = mu / pow(fabs(nu), 1 / k);
				pops[i].code[j] = step * fabs(pops[i].code[j] - pops[i].pbest_code[j]) 
								+ alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			}
			else if(rand() < RAND_MAX / 2)
				pops[i].code[j] = c + alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			else
				pops[i].code[j] = c - alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);

			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}


int compare_fit_dec(const void *p1, const void *p2)
{
	struct particle *tsk1, *tsk2;
	tsk1 = (struct particle *)p1;
	tsk2 = (struct particle *)p2;
	return (int)(tsk2->fit - tsk1->fit);
}
void sort_fit_dec(struct particle *swarm, int num_pcls)
{
	qsort(swarm, num_pcls, sizeof(struct particle), compare_fit_dec);
}


//Guolong Yu, Yong Zhao, Zhongwei Cui, and Yu Zuo
//A QPSO algorithm based on hierarchical weight and its application in cloud computing task scheduling
//Computer Science and Information Systems 2021 Volume 18, Issue 1, Pages: 189-212
//https://doi.org/10.2298/CSIS200223033Y
double hwqpos_update(double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai, c, ln_u;
	double delta[MAX_POP];

	//hwqpos_avg_pb(avg_pb);
	sort_fit_dec(pops, num_pop);
	for(i=0; i<num_pop; i++)
	{
		if(i < 0.1 * num_pop)
			delta[i] = 0.1 * rand() / (RAND_MAX + 1.0) + 1.4;
		else if(i < 0.3 * num_pop)
			delta[i] = 0.2 * rand() / (RAND_MAX + 1.0) + 1.2;
		else if(i < 0.6 * num_pop)
			delta[i] = 0.3 * rand() / (RAND_MAX + 1.0) + 0.9;
		else
			delta[i] = 0.4 * rand() / (RAND_MAX + 1.0) + 0.5;
	}
	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pop; i++)
		{
			tmp += delta[i] * pops[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pop; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai = 1.0 * rand() / RAND_MAX;
			c = fai * pops[i].pbest_code[j] + (1 - fai) * gbest_code[j];
			ln_u = log(1.0 * RAND_MAX / rand());
			if(rand() < RAND_MAX / 2)
				pops[i].code[j] = c + alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);
			else
				pops[i].code[j] = c - alpha * ln_u * fabs(pops[i].code[j] - avg_pb[j]);

			//if code_min != 0, this must be revised.
			pops[i].code[j] = fabs((int)(pops[i].code[j] / code_max[j]) * code_max[j] - pops[i].code[j]);
		}
		pops[i].fit = best =  1.0 * cal_fit_core(pops[i].code);
		if(pops[i].pbest < best)
		{
			pops[i].pbest = best;
			memcpy(pops[i].pbest_code, pops[i].code, num_task * sizeof(double));
		}
		if(pops[i].pbest > gbest)
		{
			gbest = pops[i].pbest;
			memcpy(gbest_code, pops[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}
int qpso_core(double (*update_pos)(double code_max[MAX_TASKS], double code_min[MAX_TASKS]),
		enum update_alpha update_a, double alpha_min, double alpha_max)
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	memset(code_max, 0, sizeof(double));
	memset(code_min, 0, sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
	}

	memset(pops, 0, MAX_POP * sizeof(struct particle));
	alpha = alpha_max;
	alpha_detal = alpha_max - alpha_min;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core(code_max, code_min);
	double diver;
	gettimeofday( &start, NULL);
	for(i=0; i<max_iter; i++)
	{
		switch (update_a)
		{
			case ADAPTIVE:
				diver = cal_diversity(pops, num_pop, num_task, code_max, code_min);
				if(diver < diver_low)
					alpha = alpha_min;
				else if(diver < diver_high)
					alpha = alpha_max;
				else
					alpha = alpha_min + (alpha_max - alpha_min) * i / max_iter;
				break;
			case LINEAR:
				alpha -= alpha_detal / max_iter;
				break;
			case QUADRATIC:
				alpha = alpha_max -  alpha_detal / max_iter / max_iter * i * i;
				break;
			case QUAD_DEC:
				alpha = alpha_min +  alpha_detal / max_iter / max_iter * (max_iter - i) * (max_iter - i);
				break;
			default:
				printf("Undefined update method for the contraction-expansion coefficient\n");
				return -1;
		}
		
		result = update_pos(code_max, code_min);
		fprintf(fp_convergence, "%d\t%.0lf\t%.0lf\n", i+1, gbest, result);
	}
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);

	return (int)result;
}

int compare_pbest_dec(const void *p1, const void *p2)
{
	struct particle *tsk1, *tsk2;
	tsk1 = (struct particle *)p1;
	tsk2 = (struct particle *)p2;
	return (int)(tsk2->pbest - tsk1->pbest);
}
void sort_pbest_dec(struct particle *swarm, int num_pcls)
{
	qsort(swarm, num_pcls, sizeof(struct particle), compare_pbest_dec);
}

//Situation 1: Union of best 50% of each of the improved swarms S1 and S2
struct particle *tournament(struct particle *swarm1, struct particle *swarm2, int num_pcls)
{
	sort_pbest_dec(swarm1, num_pcls);
	sort_pbest_dec(swarm2, num_pcls);
	memcpy(swarm1 + num_pcls/2, swarm2, num_pcls/2 * sizeof(struct particle));
	return swarm1;
}

void init_gbest(struct particle *pcls, int num_pcls)
{	
	gbest = pcls[0].pbest;
	memcpy(gbest_code, pcls[0].pbest_code, num_task * sizeof(double));
	int i;
	for(i=1; i<num_pcls; i++)
	{
		if(pcls[i].pbest > gbest)
		{
			gbest = pcls[i].pbest;
			memcpy(gbest_code, pcls[i].pbest_code, num_task * sizeof(double));
		}
	}
}

double tour_gpso_update(struct particle *pcls, int num_pcls, double code_max[MAX_TASKS], double code_min[MAX_TASKS])
{
	int i, j;
	double best = 0;
	double avg_pb[MAX_TASKS], tmp, fai1, fai2, c, ln_u;

	for(j=0; j<num_task; j++)
	{
		tmp = 0.0;
		for(i=0; i<num_pcls; i++)
		{
			tmp += pcls[i].pbest_code[j];
		}
		avg_pb[j] = tmp / num_task;
	}
	
	for(i=0; i<num_pcls; i++)
	{
		for(j=0; j<num_task; j++)
		{
			fai1 = gaussrand();
			fai2 = gaussrand();
			c = (fai1 * pcls[i].pbest_code[j] + fai2 * gbest_code[j]) / (fai1 + fai2);
			ln_u = log(gaussrand());
			if(rand() < RAND_MAX / 2)
				pcls[i].code[j] = c + alpha * ln_u * fabs(pcls[i].code[j] - avg_pb[j]);
			else
				pcls[i].code[j] = c - alpha * ln_u * fabs(pcls[i].code[j] - avg_pb[j]);

			//if code_min != 0, this must be revised.
			pcls[i].code[j] = fabs((int)(pcls[i].code[j] / code_max[j]) * code_max[j] - pcls[i].code[j]);
		}
		best =  1.0 * cal_fit_core(pcls[i].code);
		if(pcls[i].pbest < best)
		{
			pcls[i].pbest = best;
			memcpy(pcls[i].pbest_code, pcls[i].code, num_task * sizeof(double));
		}
		if(pcls[i].pbest > gbest)
		{
			gbest = pcls[i].pbest;
			memcpy(gbest_code, pcls[i].pbest_code, num_task * sizeof(double));
		}
	}
	return best;
}


//CJ-SoftC2020-15-A new QPSO based hybrid algorithm for constrained optimization problems via tournamenting process
int tour_qpso_core(		double (*update_pos)(struct particle *pcls, int num_pcls, 
												double code_max[MAX_TASKS], double code_min[MAX_TASKS]),
							struct particle* (*tournament)(struct particle *swarm1, struct particle *swarm2, int num_pcls),
							double alpha_min, double alpha_max)
{
	double code_max[MAX_TASKS];
	double code_min[MAX_TASKS];
	memset(code_max, 0, sizeof(double));
	memset(code_min, 0, sizeof(double));
	int i, j;
	int edge_cores = 0;
	for(i=0; i<num_edge; i++)
	{
		for(j=0; j<edges[i].num_srv; j++)
		{
			edge_cores += edges[i].srv[j].num_core;
		}
	}
	for(i=0; i<num_task; i++)
	{
		code_max[i] = tasks[i].dev->num_core + edge_cores + 1;
	}

	memset(pops, 0, MAX_POP * sizeof(struct particle));
	alpha = alpha_max;
	alpha_detal = alpha_max - alpha_min;
	struct timeval start, end;
//	double timeuse;
	double result;
	
	init_pop_core(code_max, code_min);
	struct particle *pcls[NUM_SWARM];
	//swams winning in the first group of tournament
	struct particle *winned_pcl[NUM_SWARM/GROUP_SIZE + 1];
	//swams winning in the second group of tournament
	struct particle *last_pcl;
	// the population number for each swarm
	int num_pcls = num_pop / NUM_SWARM;
	int iter;
	for(iter=0; iter < NUM_SWARM; iter++)
	{
		pcls[iter] = pops + num_pcls * iter;
		//(struct particle *)malloc(num_particles * sizeof(struct particle));
		//memcpy(particles, pops + num_particles * iter, num_particles * sizeof(struct particle));
	}
	double diver;
	gettimeofday(&start, NULL);
	for(iter = 0; iter < NUM_SWARM; iter++)
	{
		init_gbest(pcls[iter], num_pcls);
		for(i=0; i<max_iter; i++)
		{
			diver = cal_diversity(pcls[iter], num_pcls, num_task, code_max, code_min);
			if(diver < diver_low)
				alpha = alpha_min;
			else if(diver < diver_high)
				alpha = alpha_max;
			else
				alpha = alpha_min +  alpha_detal / max_iter / max_iter * (max_iter - i) * (max_iter - i);
			break;
			result = update_pos(pcls[iter], num_pcls, code_max, code_min);
			fprintf(fp_convergence, "%d\t%.0lf\t%.0lf\n", i+1, gbest, result);
		}
	}
	for(iter = 0; iter < NUM_SWARM/GROUP_SIZE; iter++)
	{
		winned_pcl[iter] = tournament(pcls[iter * GROUP_SIZE], pcls[iter * GROUP_SIZE + 1], num_pcls);
	}
	
	int last_num_swarm = NUM_SWARM/GROUP_SIZE;
	
	for(iter = 0; iter < last_num_swarm; iter++)
	{
		init_gbest(winned_pcl[iter], num_pcls);
		for(i=0; i<max_iter; i++)
		{
			diver = cal_diversity(winned_pcl[iter], num_pcls, num_task, code_max, code_min);
			if(diver < diver_low)
				alpha = alpha_min;
			else if(diver < diver_high)
				alpha = alpha_max;
			else
				alpha = alpha_min +  alpha_detal / max_iter / max_iter * (max_iter - i) * (max_iter - i);
			break;
			result = update_pos(winned_pcl[iter], num_pcls, code_max, code_min);
			fprintf(fp_convergence, "%d\t%.0lf\t%.0lf\n", i+1, gbest, result);
		}
	}
	last_pcl = tournament(winned_pcl[0], winned_pcl[1], num_pcls);

	init_gbest(last_pcl, num_pcls);
	for(i=0; i<max_iter; i++)
	{
		diver = cal_diversity(last_pcl, num_pcls, num_task, code_max, code_min);
		if(diver < diver_low)
			alpha = alpha_min;
		else if(diver < diver_high)
			alpha = alpha_max;
		else
			alpha = alpha_min +  alpha_detal / max_iter / max_iter * (max_iter - i) * (max_iter - i);
		break;
		result = update_pos(last_pcl, num_pcls, code_max, code_min);
		fprintf(fp_convergence, "%d\t%.0lf\t%.0lf\n", i+1, gbest, result);
	}
	
	gettimeofday( &end, NULL);
//	timeuse = ( end.tv_sec - start.tv_sec ) + 1.0 * (end.tv_usec - start.tv_usec) / 1000000;
//	printf("after update_pops_center: %lf seconds\n", timeuse);

	finish_size = 0.0;
	result = cal_fit_core(gbest_code);

	return (int)result;
}




